#include "itkImageRegistrationMethod.h"
#include "itkMeanSquaresImageToImageMetric.h"
#include "itkLinearInterpolateImageFunction.h"
#include "itkRegularStepGradientDescentOptimizer.h"
#include "itkImage.h"


#include "itkCenteredTransformInitializer.h"

#include "itkCenteredSimilarity2DTransform.h"
//#include "itkCenteredRigid2DTransform.h"

#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"

#include "itkResampleImageFilter.h"
#include "itkCastImageFilter.h"
#include "itkSubtractImageFilter.h"
#include "itkRescaleIntensityImageFilter.h"
#include "itkIdentityTransform.h"




int main(int argc, char **argv) {

	std::cout << "--== Starting process ==--" << std::endl;
	
	const    unsigned int    Dimension = 2;
	typedef  float           PixelType;

	typedef itk::Image< PixelType, Dimension >  FixedImageType;
	typedef itk::Image< PixelType, Dimension >  MovingImageType;

	typedef itk::CenteredSimilarity2DTransform< double > TransformType;

	typedef itk::RegularStepGradientDescentOptimizer       OptimizerType;
	typedef itk::MeanSquaresImageToImageMetric< FixedImageType, MovingImageType > MetricType;
	typedef itk::LinearInterpolateImageFunction< MovingImageType, double > InterpolatorType;
	typedef itk::ImageRegistrationMethod< FixedImageType, MovingImageType > RegistrationType;

	MetricType::Pointer         metric        = MetricType::New();
	OptimizerType::Pointer      optimizer     = OptimizerType::New();
	InterpolatorType::Pointer   interpolator  = InterpolatorType::New();
	RegistrationType::Pointer   registration  = RegistrationType::New();
	TransformType::Pointer  transform = TransformType::New();

	registration->SetMetric(metric);
	registration->SetOptimizer(optimizer);
	registration->SetTransform(transform);
	registration->SetInterpolator(interpolator);

	typedef itk::ImageFileReader< FixedImageType  > FixedImageReaderType;
	typedef itk::ImageFileReader< MovingImageType > MovingImageReaderType;

	FixedImageReaderType::Pointer  fixedImageReader  = FixedImageReaderType::New();
	MovingImageReaderType::Pointer movingImageReader = MovingImageReaderType::New();

	fixedImageReader->SetFileName(  "square.png" );
	movingImageReader->SetFileName( "lady.png" );


	registration->SetFixedImage(    fixedImageReader->GetOutput()    );
	registration->SetMovingImage(   movingImageReader->GetOutput()   );
	fixedImageReader->Update();

	registration->SetFixedImageRegion( fixedImageReader->GetOutput()->GetBufferedRegion() );

	typedef itk::CenteredTransformInitializer< 
                                    TransformType, 
                                    FixedImageType, 
                                    MovingImageType >  TransformInitializerType;

	TransformInitializerType::Pointer initializer = TransformInitializerType::New();

	initializer->SetTransform(   transform );

	initializer->SetFixedImage(  fixedImageReader->GetOutput() );
	initializer->SetMovingImage( movingImageReader->GetOutput() );

	initializer->MomentsOn();

	initializer->InitializeTransform();

	double initialScale = 1.0;
	double initialAngle = 0.0;

	transform->SetScale( initialScale );
	transform->SetAngle( initialAngle );

	registration->SetInitialTransformParameters( transform->GetParameters() );

	typedef OptimizerType::ScalesType       OptimizerScalesType;
	OptimizerScalesType optimizerScales( transform->GetNumberOfParameters() );
	const double translationScale = 1.0 / 100.0;

	optimizerScales[0] = 10.0;
	optimizerScales[1] =  1.0;
	optimizerScales[2] =  translationScale;
	optimizerScales[3] =  translationScale;
	optimizerScales[4] =  translationScale;
	optimizerScales[5] =  translationScale;

	optimizer->SetScales( optimizerScales );

	double steplength = 1.0;

	optimizer->SetMaximumStepLength( steplength ); 
	optimizer->SetMinimumStepLength( 0.0001 );
	optimizer->SetNumberOfIterations( 500 );

	registration->StartRegistration();

	OptimizerType::ParametersType finalParameters = 
                    registration->GetLastTransformParameters();

	const double finalScale           = finalParameters[0];
	const double finalAngle           = finalParameters[1];
	const double finalRotationCenterX = finalParameters[2];
	const double finalRotationCenterY = finalParameters[3];
	const double finalTranslationX    = finalParameters[4];
	const double finalTranslationY    = finalParameters[5];

	const unsigned int numberOfIterations = optimizer->GetCurrentIteration();

	const double bestValue = optimizer->GetValue();


	// Print out results
	//
	const double finalAngleInDegrees = finalAngle * 180.0 / vnl_math::pi;

	std::cout << std::endl;
	std::cout << "Result = " << std::endl;
	std::cout << " Scale         = " << finalScale  << std::endl;
	std::cout << " Angle (radians) " << finalAngle  << std::endl;
	std::cout << " Angle (degrees) " << finalAngleInDegrees  << std::endl;
	std::cout << " Center X      = " << finalRotationCenterX  << std::endl;
	std::cout << " Center Y      = " << finalRotationCenterY  << std::endl;
	std::cout << " Translation X = " << finalTranslationX  << std::endl;
	std::cout << " Translation Y = " << finalTranslationY  << std::endl;
	std::cout << " Iterations    = " << numberOfIterations << std::endl;
	std::cout << " Metric value  = " << bestValue          << std::endl;


	//////////////////////////////////////////////////////////////////////////////////////////////

	return 0;
}